/* $Id: planets.c,v 1.13 2000/02/01 21:22:18 rsmith Exp $
 * This file is part of xplanets
 * Copyright (C) 1999-2000  R.F. Smith
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * You can contact the author at the following addresses:
 *     e-mail: rsmith@xs4all.nl
 * snail-mail: R.F. Smith
 *             Dr. Hermansweg 36
 *             5624 HR Eindhoven
 *             The Netherlands
 *
 * $Log: planets.c,v $
 * Revision 1.13  2000/02/01 21:22:18  rsmith
 * w2s now exported as pl_world_to_screen
 *
 * Revision 1.12  2000/01/30 16:22:36  rsmith
 * Added planet_t.r_screen calcluation tot set_scale.
 * Adapted pl_project_to_screen to also show partially visible planets.
 *
 * Revision 1.11  2000/01/29 19:11:10  rsmith
 * Moved rel_pos and rel_speed arrays to planets.*.
 * Added pl_update_rel.
 *
 * Revision 1.10  2000/01/28 19:43:41  rsmith
 * Changed ship.dv to ship.acc.
 * Release 0.4.0
 *
 * Revision 1.9  2000/01/27 21:08:09  rsmith
 * Implemented the "delta v" steering mechanism.
 *
 * Revision 1.8  2000/01/23 21:45:20  rsmith
 * Added extra fields to planet[] initialization.
 *
 * Revision 1.7  2000/01/23 11:12:35  rsmith
 * Implemented the new view interface.
 *
 * Revision 1.6  2000/01/14 20:03:32  rsmith
 * - Implemented changes from planet.h
 * - Release 0.3.4
 *
 * Revision 1.5  2000/01/09 10:35:31  rsmith
 * Frozen for release 0.3.0
 *
 * Revision 1.4  2000/01/07 13:44:10  rsmith
 * Changed "pl_set_window_num" to "pl_set_window_size_num" on line 372.
 *
 * Revision 1.3  2000/01/07 13:42:15  rsmith
 * - Removed most "pl_init" parameters.
 * - Physics model works.
 * - Added routine: "pl_set_window_size_num", "pl_set_ship_std_orbit" and
 *   "pl_world_to_screen".
 *
 * Revision 1.2  2000/01/07 09:48:53  rsmith
 * Added "starttime" parameter to "pl_init".
 *
 * Revision 1.1  2000/01/05 21:37:42  rsmith
 * Initial revision
 *
 */

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "planets.h"

#define G 8.6e-11                       /* Newton's gravity constant */
#define AU 1.496e11                     /* Astronomical Unit [meters] */
#define AU2MTR(au) ((au)*AU)            /* convert AU to meters */
#define D2RPS(days) (2*M_PI/((days)*24.0*3600.0)) /* days to radians/second */
#define DEG2RAD(d) ((d)*M_PI/180.0)             /* degrees to radians */

/*************** EXPORTS **************/ 
ship_t ship;            /* the spaceship */
planet_t planet[NUM_PLANETS] = 
{
  {     /* the Sun */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    1.989e30,                 /* mass [kg] */
    6.96e8,                   /* planetary radius [m] */
    0.0,                      /* orbital radius [m] */
    0.0,                      /* current angle [rad] */
    0.0,                      /* angular velocity [rad/s] */
    0.0,                      /* sync. orbit angular velocity [radians/s] */
    0.0,                      /* sync. orbit radius [m] */
    "sun"                     /* name */
  },
  {     /* Mercury */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    3.303e+23,                /* mass [kg] */
    2.4335e6,                 /* planetary radius [m] */
    AU2MTR(0.387),            /* orbital radius [m] */
    DEG2RAD(281.18017),       /* current angle [rad] */
    D2RPS(87.97),             /* angular velocity [rad/s] */
    1.2400e-6,                /* sync. orbit angular velocity [radians/s] */
    2.0025e8,                 /* sync. orbit radius [m] */
    "mercury"                 /* name */
  },
  {     /* Venus */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    4.869e+24,                /* mass [kg] */
    12.119e6/2,               /* planetary radius [m] */
    AU2MTR(0.7233),           /* orbital radius [m] */
    DEG2RAD(20.17002),        /* current angle [rad] */
    D2RPS(224.70),            /* angular velocity [rad/s] */
    2.9924e-7,                /* sync. orbit angular velocity [radians/s] */
    1.2436e9,                 /* sync. orbit radius [m] */
    "venus"                   /* name */
  },
  {     /* Earth */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    5.976e+24,                /* mass [kg] */
    12.757e6/2,               /* planetary radius [m] */
    AU2MTR(1.0),              /* orbital radius [m] */
    DEG2RAD(333.58600),       /* current angle [rad] */
    D2RPS(365.256),           /* angular velocity [rad/s] */
    7.2921e-5,                /* sync. orbit angular velocity [radians/s] */
    3.5378e7,                 /* sync. orbit radius [m] */
    "earth"                   /* name */
  },
  {     /* Mars */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    6.421e+23,                /* mass [kg] */
    6.758e6/2,                /* planetary radius [m] */
    AU2MTR(1.524),            /* orbital radius [m] */
    DEG2RAD(73.77336),        /* current angle [rad] */
    D2RPS(686.98),            /* angular velocity [rad/s] */
    7.0882e-5,                /* sync. orbit angular velocity [radians/s] */
    1.8773e7,                 /* sync. orbit radius [m] */
    "mars"                    /* name */
  },
  {     /* Jupiter */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    1.900e+27,                /* mass [kg] */
    142.796e6/2,              /* planetary radius [m] */
    AU2MTR(5.203),            /* orbital radius [m] */
    DEG2RAD(292.64251),       /* current angle [rad] */
    D2RPS(4332.0),            /* angular velocity [rad/s] */
    1.7585e-4,                /* sync. orbit angular velocity [radians/s] */
    1.3249e8,                 /* sync. orbit radius [m] */
    "jupiter"                 /* name */
  },
  {     /* Saturn */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    5.688e+26,                /* mass [kg] */
    116.980e6/2,              /* planetary radius [m] */
    AU2MTR(9.522),            /* orbital radius [m] */
    DEG2RAD(8.91324),         /* current angle [rad] */
    D2RPS(10.76e3),           /* angular velocity [rad/s] */
    1.7055e-4,                /* sync. orbit angular velocity [radians/s] */
    9.0803e7,                 /* sync. orbit radius [m] */
    "saturn"                  /* name */
  },
  {     /* Uranus */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    8.686e+25,                /* mass [kg] */
    47.455e6/2,               /* planetary radius [m] */
    AU2MTR(19.28),            /* orbital radius [m] */
    DEG2RAD(298.87491),       /* current angle [rad] */
    D2RPS(30.69e3),           /* angular velocity [rad/s] */
    -9.7504e-5,               /* sync. orbit angular velocity [radians/s] */
    7.0641e7,                 /* sync. orbit radius [m] */
    "uranus"                  /* name */
  },
  {     /* Neptune */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    1.024e+26,                /* mass [kg] */
    43.118e6/2,               /* planetary radius [m] */
    AU2MTR(30.15),            /* orbital radius [m] */
    DEG2RAD(297.58472),       /* current angle [rad] */
    D2RPS(60.19e3),           /* angular velocity [rad/s] */
    1.0833e-4,                /* sync. orbit angular velocity [radians/s] */
    6.9574e7,                 /* sync. orbit radius [m] */
    "neptune"                 /* name */
  },
  {     /* Pluto */
    {0,0},{0,0},{0,0},0,0,    /* current position and speed data */
    1.27e+22,                 /* mass [kg] */
    11.5e6/2,                 /* planetary radius [m] */
    AU2MTR(39.8),             /* orbital radius [m] */
    DEG2RAD(0),               /* current angle [rad] (est.) */
    D2RPS(91.28e3),           /* angular velocity [rad/s] */
    1.1385e-5,                /* sync. orbit angular velocity [radians/s] */
    1.5814e7,                 /* sync. orbit radius [m] */
    "pluto"                   /* name */
  }
};
rel_t rel_pos[NUM_PLANETS];
rel_t rel_speed[NUM_PLANETS];
int pl_min_screen_size = 4;
cd_t winpos;     /* top left of the window, [world coordinates] */
cd_t windim = {1,1};     /* extent of the window [world coordinates] */

/*************** IMPORTS **************/


/************ STATIC DATA *************/
/* display variables */
static double scale = 1;        /* scaling factor [meter/pixel] */
static ci_t idim = {1,1};       /* screen width and height [pixels]
                                 * Initialize to non-zero value, or *
                                 * set_screen will dump core! */

/********** STATIC FUNCTIONS **********/

static void set_scale (void);
/*  static int  pl_world_to_screen (const cd_t *world, ci_t *screen); */
static void move_planet (planet_t *p, double s);
static void sim_step (double seconds);

/*********** IMPLEMENTATION ***********/

/* set_scale: Recalculate the global "scale" variable from "windin" and
 * "idim". Update the screen positions of the planets and the ship. */
void
set_scale (void)
{
  double sx, sy;
  cd_t dimdiff;
  int i;

  /* recalculate the scale in meters/pixel */
  sx = windim.x / idim.x;
  sy = windim.y / idim.y;
  scale = (sx > sy) ? sx : sy;
  /* enlarge window to follow screen aspect ratio */
  dimdiff.x = idim.x * scale - windim.x;
  dimdiff.y = idim.y * scale - windim.y;
  winpos.x -= dimdiff.x/2;
  winpos.y += dimdiff.y/2;
  windim.x += dimdiff.x;
  windim.y += dimdiff.y;
  /* recalculate the on-screen raduis */
  for (i = 0; i < NUM_PLANETS; i++) {
    planet[i].r_screen = planet[i].radius / scale;
    if (planet[i].r_screen < pl_min_screen_size)
      planet[i].r_screen = pl_min_screen_size;
  }  
}

/* pl_world_to_screen: Convert a coordinate in world space to one in screen space */
/*  int  */
/*  pl_world_to_screen (const cd_t *world, ci_t *screen) */
/*  { */
/*    cd_t tmp; */
/*    if (world == 0 || screen == 0) */
/*      return 1; */
/*    tmp.x = world->x - winpos.x; */
/*    tmp.y = -(world->y - winpos.y); */
/*    screen->x = rint (tmp.x / scale); */
/*    screen->y = rint (tmp.y / scale); */
/*    return 0; */
/*  } */

/* move_planet: Move the planet "p" along it's orbit, "s" seconds from it's
 * present position */
void 
move_planet (planet_t *p, double s)
{
  if (p == 0 || s < 0.0) {
    return;
  }
  p->angle += p->w * s;
  /* calculate new coorinates */
  p->pos.x = p->orbit * cos (p->angle);
  p->pos.y = p->orbit * sin (p->angle);
  p->speed.x = - p->w * p->orbit * sin (p->angle);
  p->speed.y =   p->w * p->orbit * cos (p->angle);
}


/* This is where the simulation is run */
void 
sim_step (double seconds)
{
  int t;
  cd_t a, dv;
  
  /* calculate new coorinates */
  a = pl_gravity (ship.pos);
  dv.x = a.x * seconds;
  dv.y = a.y * seconds;
  if (ship.engine_on) {
    dv.x += ship.acc * seconds * cos (ship.dir);
    dv.y += ship.acc * seconds * sin (ship.dir);
  }
  ship.pos.x += ship.speed.x * seconds + dv.x *seconds / 2;
  ship.pos.y += ship.speed.y * seconds + dv.y *seconds / 2;
  ship.speed.x += dv.x;
  ship.speed.y += dv.y;
  for (t = 1; t < NUM_PLANETS; t++) {
    move_planet (&planet[t], seconds);
  }
}

/* pl_init: Initiate the planets module. (Front end for pl_set_screen_size
 * and pl_set_window_size.) */
int 
pl_init (int screen_width, int screen_height)
{
  int i, rv = 0;
  time_t epoch_time, starttime;
  struct tm data_epoch; /* the provided angles in the planet array are
                         *  taken on this date */

  memset (&data_epoch, 0, sizeof(struct tm));
  data_epoch.tm_mday = 25;
  data_epoch.tm_mon = 8;
  data_epoch.tm_year = 1996;
  epoch_time = mktime (&data_epoch);
  time (&starttime);
  for (i = 0; i < NUM_PLANETS; i++) {
    move_planet (&planet[i], (double)(starttime-epoch_time));
  }
  pl_set_ship_std_orbit (3);
  rv = pl_set_view_screen (screen_width, screen_height);
  if (rv != 0) 
    return rv;
  rv = pl_set_view_size (2*planet[4].orbit,
                         2*planet[4].orbit);
  if (rv != 0) 
    return rv;
  pl_project_to_screen ();
  return rv;
}

/* pl_set_view_screen: Set the size of the display screen (in pixels). If
 * either "width" or "height" are <= 0, it does not set the size and
 * returns a non-zero value.  */
int 
pl_set_view_screen (int width, int height)
{
  if (width < 0 || height < 0) {
    return 1;
  }
  idim.x = width;
  idim.y = height;
  set_scale ();
  return 0;
}

/* pl_set_view_center: Set the point in world space where the view is
 * centred. */
void 
pl_set_view_center (double x, double y)
{
  winpos.x = x - windim.x / 2;  /* left */
  winpos.y = y + windim.y / 2;  /* top */
}

/* pl_set_view_size: Set the view window size in world coordinates. */
int 
pl_set_view_size (double width, double height)
{
  cd_t cen;

  if (width <= 0 || height <= 0) {
    return 1;
  }
  cen.x = winpos.x + windim.x / 2;
  cen.y = winpos.y - windim.y / 2;
  windim.x = width;
  windim.y = height;
  winpos.x = cen.x - windim.x / 2;
  winpos.y = cen.y + windim.y / 2;
  set_scale ();
  return 0;
}

/* pl_sim_step: Calculate the new positions at "seconds" in the future prom
 * the current position. */
void 
pl_sim_step (unsigned seconds)
{
  unsigned max_step = 3600*5, quot, rem, t;

  /* do multiple steps if seconds > max_step, to avoid numerical
   * instabilities */
  quot = seconds/max_step;
  rem = seconds-quot*max_step;
  for (t = 0; t < quot; t++) {
    sim_step ((double)max_step);
  }
  if (rem)
    sim_step ((double)rem);
}

/* pl_project_to_screen: Calculate the screen coordinates. */
void 
pl_project_to_screen (void)
{
  int t;

  for (t = 0; t < NUM_PLANETS; t++) {
    pl_world_to_screen (&planet[t].pos, &planet[t].screen);
    if (planet[t].screen.x < -planet[t].r_screen
        || planet[t].screen.x > idim.x + planet[t].r_screen
        || planet[t].screen.y < -planet[t].r_screen
        || planet[t].screen.y > idim.y + planet[t].r_screen) {
      planet[t].on_screen = 0;
    } else {
      planet[t].on_screen = 1;
    }
  }
  pl_world_to_screen (&ship.pos, &ship.screen);
  if (ship.screen.x < 0 || ship.screen.x > idim.x
      || ship.screen.y < 0 || ship.screen.y > idim.y) {
    ship.on_screen = 0;
  } else {
    ship.on_screen = 1;
  }
  
}

void 
pl_update_rel (void)
{
  int i;
  
  /* compute new relative position/speed */
  for (i = 0; i < NUM_PLANETS; i++) {
    pl_rel (&planet[i].speed, &ship.speed, &rel_speed[i]);
    pl_rel (&planet[i].pos, &ship.pos, &rel_pos[i]);
    rel_pos[i].len -= planet[i].radius;
  }
}

void 
pl_set_ship_std_orbit (int t)
{
#define ORBIT 10
  double v;

  if (t < 0 || t >= NUM_PLANETS)
    return;
  v = sqrt(G*planet[t].mass/(ORBIT*planet[t].radius));
  ship.pos.x = planet[t].pos.x + ORBIT*planet[t].radius;
  ship.pos.y = planet[t].pos.y;
  ship.speed.x = planet[t].speed.x;
  ship.speed.y = planet[t].speed.y + v;
  ship.engine_on = 0;
#undef ORBIT
}

int pl_world_to_screen (const cd_t *world, ci_t *screen)
{
  cd_t tmp;

  if (world == 0 || screen == 0)
    return 1;
  /* convert to screen coordinate system */
  tmp.x = world->x - winpos.x;
  tmp.y = -(world->y - winpos.y);       /* in screen coord. is y down! */
  /* scale to pixel coordinates */
  screen->x = rint (tmp.x / scale);
  screen->y = rint (tmp.y / scale);
  return 0;
}

/* pl_gravity: Calculate the gravity accelleration vector at a given point
 * in world space. This assumes that all planets are point-masses, so even
 * a position inside a planet will produce a result. */
cd_t 
pl_gravity (cd_t pos) 
{
  int t;
  cd_t a = {0.0,0.0};

  for (t = 0; t < NUM_PLANETS; t++) {
    double aa, d;
    cd_t d2;
    pl_distance (&pos, &planet[t].pos, &d, &d2);
      aa = planet[t].mass * G /(d*d);
      a.x += d2.x * aa;
      a.y += d2.y * aa;
  }
  return a;
}

/* pl_distance: Calculate the distance between the positions pointed to by
 * "a" and "b", return the value in the variable pointed to by "res". A
 * normalized vector a->b is returned in "res2".  If any of the parameters
 * "a", "b" or "res" are NULL it returns a non-zero value. Normally it
 * returns 0. */
int 
pl_distance (const cd_t *a, const cd_t *b, double *res, cd_t *res2) 
{ 
  cd_t d;
  double len;

  if (a == 0 || b == 0 || res == 0) {
    return 1;
  }
  d.x = (b->x - a->x); 
  d.y = (b->y - a->y);
  len = sqrt (d.x*d.x+d.y*d.y);
  *res = len;
  if (res2) {
    res2->x = d.x/len;
    res2->y = d.y/len;    
  }
  return 0;
}

int 
pl_rel (const cd_t *a, const cd_t *b, rel_t *res)
{
  double dx, dy;
 
  if (a == 0 || b == 0 || res == 0) {
    return 1;
  }
  dx = a->x - b->x;
  dy = a->y - b->y;
  res->len = sqrt (dx*dx + dy*dy);
  res->phi = atan2 (dy, dx);
  return 0;
}
